﻿using InABox.Clients;
using InABox.Configuration;
using InABox.Core.Postable;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace InABox.Core
{

    public interface IPosterEngine<TPostable>
        where TPostable : Entity, IPostable, IRemotable, IPersistent, new()
    {
        bool Process(IDataModel<TPostable> model);
    }

    public interface IPosterEngine<TPostable, TPoster, TSettings> : IPosterEngine<TPostable>
        where TPostable : Entity, IPostable, IRemotable, IPersistent, new()
        where TPoster : IPoster<TPostable, TSettings>
        where TSettings : PosterSettings, new()
    {
    }

    /// <summary>
    /// A base class for all <see cref="IPosterEngine{TPostable}"/>. A concrete instance of this will be loaded by
    /// <see cref="PosterUtils.Process{T}(IDataModel{T})"/>; a new instance is guaranteed to be created each time that method is called.
    /// </summary>
    /// <typeparam name="TPostable"></typeparam>
    /// <typeparam name="TPoster"></typeparam>
    /// <typeparam name="TSettings"></typeparam>
    public abstract class PosterEngine<TPostable, TPoster, TSettings> : IPosterEngine<TPostable, TPoster, TSettings>
        where TPostable : Entity, IPostable, IRemotable, IPersistent, new()
        where TPoster : IPoster<TPostable, TSettings>
        where TSettings : PosterSettings, new()
    {
        protected static TPoster Poster = GetPoster();

        private static Type[]? _posters;

        private static TPoster GetPoster()
        {
            _posters ??= CoreUtils.TypeList(
                AppDomain.CurrentDomain.GetAssemblies(),
                x => x.IsClass
                    && !x.IsAbstract
                    && !x.IsGenericType
                    && x.HasInterface(typeof(IPoster<,>))
            ).ToArray();

            var type = _posters.Where(x => typeof(TPoster).IsAssignableFrom(x)).FirstOrDefault()
                ?? throw new Exception($"No poster of type {typeof(TPoster)}.");
            return (TPoster)Activator.CreateInstance(type);
        }

        protected static TSettings GetSettings()
        {
            return PosterUtils.LoadPosterSettings<TPostable, TSettings>();
        }

        protected static void SaveSettings(TSettings settings)
        {
            PosterUtils.SavePosterSettings<TPostable, TSettings>(settings);
        }

        /// <summary>
        /// Returns the <see cref="TSettings.Script"/>, if <see cref="TSettings.ScriptEnabled"/> is <see langword="true"/>;
        /// otherwise, returns <see langword="null"/>.
        /// </summary>
        protected static string? GetScript()
        {
            var settings = GetSettings();
            return settings.ScriptEnabled ? settings.Script : null;
        }

        protected abstract bool DoProcess(IDataModel<TPostable> model);

        /// <summary>
        /// Process the <paramref name="model"/> before loading;
        /// </summary>
        /// <param name="model"></param>
        /// <returns><see langword="false"/> if the processing must be cancelled.</returns>
        public abstract bool BeforePost(IDataModel<TPostable> model);

        /// <summary>
        /// Prior to saving the <typeparamref name="TPostable"/> entities, make any necessary changes to those entities.
        /// This is only called if <see cref="Process(IDataModel{TPostable})"/> returned <see langword="true"/>.
        /// </summary>
        /// <param name="model"></param>
        public abstract void AfterPost(IDataModel<TPostable> model);

        public bool Process(IDataModel<TPostable> model)
        {
            if (!BeforePost(model))
            {
                return false;
            }
            model.LoadModel();

            var data = model.GetTable<TPostable>();

            if (!data.Rows.Any())
            {
                throw new EmptyPostException();
            }
            if(data.Rows.Any(x => x.Get<TPostable, PostedStatus>(x => x.PostedStatus) == PostedStatus.Posted))
            {
                throw new RepostedException();
            }

            try
            {
                var success = DoProcess(model);
                if (success)
                {
                    AfterPost(model);
                }

                var entities = data.ToObjects<TPostable>().ToList();

                if (success)
                {
                    foreach (var post in entities)
                    {
                        post.Posted = DateTime.Now;
                        post.PostedStatus = PostedStatus.Posted;
                    }
                    new Client<TPostable>().Save(entities, "Posted by user.");
                }
                else
                {
                    foreach (var post in entities)
                    {
                        post.PostedStatus = PostedStatus.PostFailed;
                    }
                    new Client<TPostable>().Save(entities, "Post failed by user.");
                }
                return success;
            }
            catch(Exception e)
            {
                Logger.Send(LogType.Error, "", $"Post Failed: {CoreUtils.FormatException(e)}");

                var entities = data.ToObjects<TPostable>().ToList();
                foreach (var post in entities)
                {
                    post.PostedStatus = PostedStatus.PostFailed;
                }
                new Client<TPostable>().Save(entities, "Post failed by user.");
                throw;
            }
        }
    }
}
